/*
 * gtk+-immodule
 */
#include <gtk/gtk.h>
#include <gtk/gtkimcontext.h>
#include <gtk/gtkimmodule.h>
#include <gdk/gdkkeysyms.h>
#include <gdk/gdkx.h>

#include <string.h>
#include <stdlib.h>
#include "uim.h"

#define NR_CANDIDATES 8

struct preedit_segment {
  int attr;
  char *str;
};

struct candidate_window {
  GtkWidget *top_win;
  GtkWidget *clist;
  int nr_candidates;
  int candidate_index;
  int layout_begin;
  int is_active;
  int select_count;
};

typedef struct _IMUIMContext {
  /**/
  struct _GtkIMContext parent;
  struct _GtkIMContext *slave;
  uim_context uc;
  struct candidate_window cwin;
  int nr_psegs;
  struct preedit_segment *pseg;
  /**/
  GtkWidget *stat_win;
  GtkWidget *menu;
  GdkWindow *win;
  int mode;
  /**/
  struct _IMUIMContext *prev, *next;
} IMUIMContext;

static IMUIMContext context_list;

static GObjectClass *parent_class;

typedef struct _IMContextUIMClass 
{
  GtkIMContextClass parent_class;
} IMContextUIMClass;

static void im_uim_class_init (GtkIMContextClass *class);

static const GTypeInfo class_info = {
  sizeof(IMContextUIMClass),
  (GBaseInitFunc) NULL,
  (GBaseFinalizeFunc) NULL,
  (GClassInitFunc) im_uim_class_init,
  NULL,
  NULL,
  sizeof(IMUIMContext),
  0,
  (GtkObjectInitFunc) NULL,
};

static GType type_im_uim;
#define IM_UIM_CONTEXT(obj) (GTK_CHECK_CAST((obj),type_im_uim,IMUIMContext))
static const GtkIMContextInfo im_uim_info = {
  "uim",
  "UIM()",
  "gtk+",
  "",
  ""
};

static
void
im_uim_commit_string(void *ptr, char *str)
{
  IMUIMContext *uic = (IMUIMContext *)ptr;
  g_signal_emit_by_name(uic, "commit", strdup(str));
}

static void
clear_cb(void *ptr)
{
  IMUIMContext *uic = ptr;
  int i;
  for (i = 0; i < uic->nr_psegs; i++) {
    free(uic->pseg[i].str);
  }
  free(uic->pseg);
  uic->pseg = 0;
  uic->nr_psegs = 0;
}

static void
pushback_cb(void *ptr, int attr, char *str)
{
  IMUIMContext *uic = ptr;
  uic->pseg = realloc(uic->pseg,
		      sizeof(struct preedit_segment) *
		     (uic->nr_psegs + 1));
  uic->pseg[uic->nr_psegs].str = strdup(str);
  uic->pseg[uic->nr_psegs].attr = attr;
  uic->nr_psegs ++;
}

static void
update_cb(void *ptr)
{
  IMUIMContext *uic = ptr;
  g_signal_emit_by_name(uic, "preedit_changed");
}

static void
mode_cb(void *ptr, int mode)
{
  IMUIMContext *uic = ptr;  
  if (uic->mode == mode) {
    return ;
  }
  gtk_option_menu_set_history(GTK_OPTION_MENU(uic->menu), mode);
  uic->mode = mode;
}

static int
convert_keyval(int key)
{
  switch (key) {
  case GDK_BackSpace: return UKey_Backspace;
  case GDK_Delete: return UKey_Delete;
  case GDK_Escape: return UKey_Escape;
  case GDK_Return: return UKey_Return;
  case GDK_Left: return UKey_Left;
  case GDK_Up: return UKey_Up;
  case GDK_Right: return UKey_Right;
  case GDK_Down: return UKey_Down;
  }
  return key;
}

static int
convert_modifier(int mod)
{
  int rv = 0;
  if (mod & GDK_SHIFT_MASK) {
    rv |= UKey_Shift;
  }
  if (mod & GDK_CONTROL_MASK) {
    rv |= UKey_Control;
  }
  if (mod & GDK_MOD1_MASK) {
    rv |= UKey_Alt;
  }
  return rv;
}


/*
 * KEY EVENT HANDLER
 */
static gboolean
filter_keypress(GtkIMContext *ic,
		GdkEventKey *key)
{
  int rv;
  IMUIMContext *uic = IM_UIM_CONTEXT(ic);
  int kv = convert_keyval(key->keyval);
  int mod = convert_modifier(key->state);

  if (key->type == GDK_KEY_RELEASE) {
    rv = uim_release_key(uic->uc, kv, mod);
  } else {
    rv = uim_press_key(uic->uc, kv, mod);
  }
  if (rv) {
    return gtk_im_context_filter_keypress(uic->slave, key);
  }
  return TRUE;
}

static char *
get_preedit_segment(struct preedit_segment *ps,
		    PangoAttrList *attrs,
		    char *str)
{
  PangoAttribute *attr;
  if (attrs) {
    int begin, end;
    begin = strlen(str);
    end = begin + strlen(ps->str);
    if (ps->attr & UPeAttr_UnderLine) {
      attr = pango_attr_underline_new(PANGO_UNDERLINE_SINGLE);
      attr->start_index = begin;
      attr->end_index = end;
      pango_attr_list_change(attrs, attr);
    }
    if (ps->attr & UPeAttr_Reverse) {
        attr = pango_attr_foreground_new (0xffff, 0xffff, 0xffff);
        attr->start_index = begin;
        attr->end_index = end;
        pango_attr_list_change (attrs, attr);

        attr = pango_attr_background_new (0, 0, 0);
        attr->start_index = begin;
        attr->end_index = end;
        pango_attr_list_change (attrs, attr);
    }
  }
  str = (char *)realloc(str, strlen(str) + strlen(ps->str)+1);
  strcat(str, ps->str);
  return str;
}


static void
im_uim_get_preedit_string(GtkIMContext *ic, gchar **str,
			  PangoAttrList **attrs,
			  gint *cursor_pos)
{
  char *tmp;
  int i, pos = 0, cur_pos = 0;
  IMUIMContext *uic = IM_UIM_CONTEXT(ic);

  if (attrs) {
    *attrs = pango_attr_list_new();
  }
  tmp = strdup("");
  for (i = 0; i < uic->nr_psegs; i++) {
    cur_pos += strlen(tmp);
    if (uic->pseg[i].attr & UPeAttr_Cursor) {
      pos = cur_pos;
    }
    if (attrs) {
      tmp = get_preedit_segment(&uic->pseg[i], *attrs, tmp);
    } else {
      tmp = get_preedit_segment(&uic->pseg[i], 0, tmp);
    }
  }
  if (cursor_pos) {
    *cursor_pos = pos;
  }
  if (str) {
    *str = tmp;
  } else {
    free(tmp);
  }
}

static void
im_uim_commit_cb(GtkIMContext *ic,
		 const gchar *str,
		 IMUIMContext *is)
{
  g_signal_emit_by_name(is, "commit", str);
}

static void
update_status_window(IMUIMContext *uic)
{
  if (uic->win && uic->stat_win) {
    gint x,y,w,h,d;
    gdk_window_get_geometry(uic->win, &x, &y,&w,&h,&d);
    gdk_window_get_origin(uic->win, &x, &y);
    gtk_window_move(GTK_WINDOW(uic->stat_win), x+w , y+h);
  }
  if(uic->cwin.top_win && uic->cwin.is_active){
    gtk_widget_show(uic->cwin.top_win);
  }
}

void focus_in(GtkIMContext *ic)
{
  IMUIMContext *uic = IM_UIM_CONTEXT(ic);
  IMUIMContext *cc;
  for (cc = context_list.next; cc != &context_list; cc = cc->next) {
    if (cc != uic) {
      if (uic->stat_win) {
	gtk_widget_hide(uic->stat_win);
      }
      if (uic->cwin.top_win) {
	gtk_widget_hide(uic->cwin.top_win);
      }
    }
  }
  update_status_window(uic);
  if (uic->stat_win) {
    gtk_widget_show(uic->stat_win);
  }
}

void focus_out(GtkIMContext *ic)
{
  IMUIMContext *uic = IM_UIM_CONTEXT(ic);
  if (uic->win && uic->stat_win) {
    gtk_widget_hide(uic->stat_win);
  }
  if(uic->cwin.top_win){
    gtk_widget_hide(uic->cwin.top_win);
  }
}

static GdkFilterReturn
resize_window_cb (GdkXEvent *xevent, GdkEvent *ev, gpointer data)
{
  IMUIMContext *uic = IM_UIM_CONTEXT(data);
  update_status_window(uic);
  return GDK_FILTER_CONTINUE;
}


static void
set_client_window(GtkIMContext *ic, GdkWindow *w)
{
  IMUIMContext *uic = IM_UIM_CONTEXT(ic);
  if (uic->win != w) {
    if (uic->win) {
      gdk_window_remove_filter(uic->win, resize_window_cb, uic);
    }
    if (w) {
      GdkEventMask mask;
      mask = gdk_window_get_events(w);
      gdk_window_set_events(w, mask | GDK_STRUCTURE_MASK);
      gdk_window_add_filter(w, resize_window_cb, uic);
    }
  }
  uic->win = w;
  update_status_window(uic);
}


/*
 * DESTRUCTOR
 */
static void
im_uim_finalize(GObject *obj)
{
  IMUIMContext *uic = IM_UIM_CONTEXT(obj);
  set_client_window(GTK_IM_CONTEXT(uic), NULL);
  uic->next->prev = uic->prev;
  uic->prev->next = uic->next;

  if (uic->menu) {
    gtk_widget_destroy(uic->menu);
  }
  if (uic->stat_win) {
    gtk_widget_destroy(uic->stat_win);
  }
  if (uic->cwin.top_win) {
    gtk_widget_destroy(uic->cwin.top_win);
  }
  uic->stat_win = NULL;
  uic->menu = NULL;

  uim_release_context(uic->uc);

  g_signal_handlers_disconnect_by_func(uic->slave,
				       im_uim_commit_cb,
				       uic);
  g_object_unref(uic->slave);
  parent_class->finalize (obj);
}

static void
im_uim_class_init (GtkIMContextClass *class)
{
  GObjectClass *object_class = G_OBJECT_CLASS(class);
  parent_class = g_type_class_peek_parent(class);
  class->set_client_window = set_client_window;
  class->filter_keypress = filter_keypress;
  class->get_preedit_string = im_uim_get_preedit_string;
  class->focus_in = focus_in;
  class->focus_out = focus_out;
  object_class->finalize = im_uim_finalize;
}

static gint
select_mode(GtkWidget *w, gpointer p)
{
  IMUIMContext *is = p;
  int nth = gtk_option_menu_get_history(GTK_OPTION_MENU(is->menu));
  uim_set_mode(is->uc, nth);
  return 0;
}

static void
update_status_menu(IMUIMContext *uic)
{
  int i, nr;
  GtkWidget *menu;

  nr = uim_get_nr_modes(uic->uc);
  if (!nr) {
    uic->menu = NULL;
    uic->stat_win = NULL;
    return ;
  }
  uic->menu = gtk_option_menu_new();
  uic->stat_win = gtk_window_new(GTK_WINDOW_POPUP);
  menu = gtk_menu_new();
  for (i = 0; i < nr; i++) {
    GtkWidget *item;
    char *name;
    name = uim_get_mode_name(uic->uc, i);
    item = gtk_menu_item_new_with_label(name);
    gtk_menu_append(menu, item);
    gtk_signal_connect(GTK_OBJECT(item), "activate",
		       GTK_SIGNAL_FUNC(select_mode), uic);
    gtk_widget_show(item);
  }
  gtk_option_menu_set_menu(GTK_OPTION_MENU(uic->menu), menu);
  gtk_container_add(GTK_CONTAINER(uic->stat_win), uic->menu);
  gtk_option_menu_set_history(GTK_OPTION_MENU(uic->menu),
			      uim_get_current_mode(uic->uc));
  gtk_widget_show(uic->menu);
}

static void
update_mode_list_cb(void *ptr)
{
  IMUIMContext *uic = ptr;
  if (uic->menu) {
    gtk_widget_destroy(uic->menu);
  }
  if (uic->stat_win) {
    gtk_widget_destroy(uic->stat_win);
  }
  uic->menu = NULL;
  uic->stat_win = NULL;
  update_status_menu(uic);
}


static void
select_candidate_cb(GtkWidget *clist,
		    gint row, gint column,
		    GdkEventButton *event,
		    gpointer data)
{
  IMUIMContext *uic = IM_UIM_CONTEXT(data);
  if (uic->cwin.select_count == 0) {
    int idx = row + uic->cwin.layout_begin;
    uim_set_candidate_index(uic->uc, idx);
    uic->cwin.candidate_index = idx;
  } else {
    uic->cwin.select_count --;
  }
}

static void
init_candidate_win(IMUIMContext *uic)
{
  gchar *titles[2] = {"index", "Candidates"};
  struct candidate_window *cwin = &uic->cwin;
  if (cwin->top_win) {
    return ;
  }
  cwin->top_win = gtk_window_new(GTK_WINDOW_POPUP);
  cwin->clist = gtk_clist_new_with_titles(2, titles);
  cwin->is_active = 0;
  g_signal_connect(G_OBJECT(cwin->clist), "select_row",
		   G_CALLBACK(select_candidate_cb),
		   uic);
  gtk_container_add(GTK_CONTAINER(cwin->top_win), cwin->clist);
  {
    gint x,y,w,h,d;
    gdk_window_get_geometry(uic->win, &x, &y,&w,&h,&d);
    gdk_window_get_origin(uic->win, &x, &y);
    gtk_window_move(GTK_WINDOW(cwin->top_win), x , y+h);
  }
  gtk_widget_show(cwin->clist);
}

static int
trim_index(IMUIMContext *uic)
{
  struct candidate_window *cwin = &uic->cwin;
  int changed = 0;
  while (cwin->layout_begin + NR_CANDIDATES <= cwin->candidate_index) {
    changed = 1;
    cwin->layout_begin += NR_CANDIDATES;
  }  
  while (cwin->layout_begin > cwin->candidate_index) {
    changed = 1;
    cwin->layout_begin -= NR_CANDIDATES;
  }
  if (cwin->layout_begin < 0) {
    changed = 1;
    cwin->layout_begin = 0;
  }
  return changed;
}

static void
layout_candidate(IMUIMContext *uic)
{
  char *buf[2];
  int i;
  trim_index(uic);
  gtk_clist_clear(uic->cwin.clist);
  for (i = 0; i < 8; i++) {
    buf[0] = "*";
    if (i < uic->cwin.nr_candidates) {
      buf[1] = uim_get_candidate(uic->uc, i + uic->cwin.layout_begin);
    } else {
      buf[1] = "*";
    }
    gtk_clist_append(uic->cwin.clist, buf);
  }
  uic->cwin.select_count ++;
  gtk_clist_select_row(uic->cwin.clist,
		       uic->cwin.candidate_index - uic->cwin.layout_begin,
		       0);
}

static void
cand_begin_cb(void *ptr, int nr, int index)
{
  IMUIMContext *uic = ptr;
  init_candidate_win(uic);
  uic->cwin.is_active = 1;
  uic->cwin.nr_candidates = nr;
  uic->cwin.candidate_index = index;
  uic->cwin.layout_begin = 0;
  layout_candidate(uic);
  gtk_widget_show(uic->cwin.top_win);
}

static void
cand_update_cb(void *ptr, int index)
{
  IMUIMContext *uic = ptr;
  uic->cwin.candidate_index = index;
  layout_candidate(uic);
}

static void
cand_end_cb(void *ptr)
{
  IMUIMContext *uic = ptr;
  gtk_widget_hide(uic->cwin.top_win);
  uic->cwin.is_active = 0;
}


/*
 * CONSTRUCTOR
 */
GtkIMContext *
im_module_create (const gchar *context_id)
{
  GtkObject *obj = g_object_new(type_im_uim, NULL);
  IMUIMContext *uic = IM_UIM_CONTEXT(obj);
  uic->win = NULL;
  uic->menu = NULL;
  uic->stat_win = NULL;
  uic->pseg = 0;
  uic->nr_psegs = 0;

  uic->cwin.top_win = NULL;
  uic->cwin.clist = NULL;
  uic->cwin.select_count = 0;


  if (!strncmp(context_id, "uim-", 4)) {
    context_id = &context_id[4];
  }
  uic->uc =  uim_create_context(uic, "UTF-8",
				NULL, (char *)context_id,
				im_uim_commit_string);
  uim_set_mode_cb(uic->uc, mode_cb);
  uim_set_preedit_cb(uic->uc, clear_cb,
		     pushback_cb, update_cb);
  uim_set_mode_list_update_cb(uic->uc,
			      update_mode_list_cb);
  uim_set_candidate_cb(uic->uc,
		       cand_begin_cb,
		       cand_update_cb,
		       cand_end_cb);
  
  uic->mode = uim_get_current_mode(uic->uc);
  uic->slave = g_object_new(GTK_TYPE_IM_CONTEXT_SIMPLE,
			    NULL);
  g_signal_connect(G_OBJECT(uic->slave), "commit",
		   G_CALLBACK(im_uim_commit_cb), uic);
  update_status_menu(uic);
  /**/
  uic->next = context_list.next;
  uic->prev = (IMUIMContext *)&context_list;
  context_list.next->prev = uic;
  context_list.next = uic;
  return GTK_IM_CONTEXT(uic);
}

void
im_module_list(GtkIMContextInfo ***contexts,
	       int *n_contexts)
{
  uim_context uc;
  int nr, i;
  static GtkIMContextInfo **info_list;

  uim_init();
  uc = uim_create_context(NULL, "UTF-8",
			  NULL, NULL, NULL);
  nr = uim_get_nr_im(uc);

  info_list = malloc(sizeof(GtkIMContextInfo*)*nr);
  for (i = 0; i < nr; i++) {
    char *name = uim_get_im_name(uc, i);
    char *s = alloca(strlen(name) + 5);
    sprintf(s, "uim-%s", name);
    free(name);
    info_list[i] = malloc(sizeof(GtkIMContextInfo));
    info_list[i]->context_id = strdup(s);
    info_list[i]->context_name = strdup(s);
    info_list[i]->domain = "gtk+";
    info_list[i]->domain_dirname = "";
    info_list[i]->default_locales = uim_get_im_language(uc, i);
  }
  uim_release_context(uc);
  *contexts = info_list;
  *n_contexts = nr;
}

void
im_module_init(GTypeModule *type_module)
{
  uim_init();
  context_list.next = (IMUIMContext *)&context_list;
  context_list.prev = (IMUIMContext *)&context_list;
  type_im_uim =
    g_type_module_register_type(type_module,
				GTK_TYPE_IM_CONTEXT,
				"GtkIMContextUIM",
				&class_info, 0);

  
}

void
im_module_exit()
{
}
